import ij.IJ;
import ij.io.OpenDialog;
import ij.plugin.PlugIn;
import java.io.*;
import ij.ImagePlus;
import ij.ImageStack;
import ij.process.ImageProcessor;
import ij.WindowManager;
import ij.gui.GenericDialog;
import Jama.Matrix;
import Fx_Maths_PTM.CLinAlg;
import java.util.Vector;
import java.awt.Checkbox;


/**
 * Calcule la carte des normales en fonction
 * - d'un imagestack de n images
 * - d'un fichier de n position de lampes correspondant au stack
 * @author Vincent Vansuyt
 */
public class Normal_map_computing_using_PTM implements PlugIn {
	
	
	/**
	 * Convertit une chaine en double
	 * @param s
	 * @return nValeur, un double
	 */
	private double convert_to_double( String s ) {
		double nValeur = 0.0;
		try {
			nValeur = Double.valueOf(s.trim()).doubleValue();
		}
		catch (NumberFormatException nfe) {
			IJ.log("NumberFormatException: " + nfe.getMessage());
		}
		return nValeur;
	}// convert_to_double
	
	
	/**
	 * Point d'entrée du programme
	 * @param arg : aucun argument
	 */
	@Override
	public void run(String arg) {
		IJ.log("\\Clear");
		IJ.log("------- Normal_map_computing_using_PTM -- start ------");
		
		// Vérification qu'il y a au moins une image ouverte
		ImagePlus imp;
		imp = WindowManager.getCurrentImage();
		if (imp == null) {
			IJ.noImage();
			return;
		}
		// - doit être stack
		int nNb_slices = imp.getNSlices();
		if ( nNb_slices <= 1 ) {
			IJ.showMessage( "Warning", "Image must contain more than one stack" );
			return;
		}
		// Vérifications que l'image est en niveau de gris
		int nType = imp.getType();
		if ( nType != ImagePlus.GRAY8 &&
			 nType != ImagePlus.GRAY16 && 
			 nType != ImagePlus.GRAY32 ) {
			IJ.showMessage( "Warning", "Image must be gray scale");
			return;
		}
		
		OpenDialog dialog = new OpenDialog( "Select lamp position file" , "" );
		String strChemin = dialog.getDirectory();
		String strFichier_positions_lampes = dialog.getFileName();
		String strNom_complet = strChemin + "\\" + strFichier_positions_lampes;
		//IJ.log( "-> " + strNom_complet );
		
		//lecture du fichier texte et stockage dans une matrice
		//- comptage du nombre de lignes
		int nNb_lignes = 0;
		try{
			InputStream ips = new FileInputStream( strNom_complet ); 
			InputStreamReader ipsr=new InputStreamReader( ips );
			BufferedReader br = new BufferedReader( ipsr );
			String strLigne;
			String tabStr[];
			double x, y, z;
			while ((strLigne = br.readLine()) != null ){
				++nNb_lignes;
			}
			br.close(); 
		}		
		catch (Exception e){
			IJ.log( "Erreur" );
			IJ.log( e.toString());
		}
		
		//- relecture et stockage dans une matrice
		IJ.log( nNb_lignes + " lignes lues");
		Matrix vPos_lampes = new Matrix( nNb_lignes, 3, 0.0 );
		int i = 0;
		try{
			InputStream ips = new FileInputStream( strNom_complet ); 
			InputStreamReader ipsr = new InputStreamReader( ips );
			BufferedReader br = new BufferedReader( ipsr );
			String strLigne;
			String tabStr[];
			double x, y, z;
			while ((strLigne = br.readLine()) != null ){
				tabStr = strLigne.split(" ");
				
				x = convert_to_double( tabStr[0]);
				y = convert_to_double( tabStr[1]);
				z = convert_to_double( tabStr[2]);
				vPos_lampes.set( i, 0, x );
				vPos_lampes.set( i, 1, y );
				vPos_lampes.set( i, 2, z );
				++i;
			}
			br.close();
		}
		catch (Exception e){
			IJ.log( "Erreur" );
			IJ.log( e.toString());
		}
		
		// Vérification que le nombre de position est égal au nombre de
		//  slices
		if ( nNb_lignes != nNb_slices ) {
			String strM = "Error : number of slices must be equal to number of lamp position.";
			IJ.log( strM );
			IJ.showMessage( strM );
		}
		else {
			IJ.log("Nb lamps positions  == nb slices : OK");
		}
		
		// Demande a l'utilisateur de saisir les valeurs :
		//  ymin et ymax, la taille réelle de l'image
		//  xmin et xmax, la taille réelle de l'image
		//  La taille de l'image va de 0 à la largeur-hauteur de l'image
		GenericDialog gd = new GenericDialog ("Normal map computing" );
		gd.addMessage( "Texture real size" ) ;
		final int ncNb_decimales = 2;
		double x_min = 0.0, x_max = 1.0;
		double y_min = 0.0, y_max = 1.0;
		gd.addNumericField( "x min :" , x_min , ncNb_decimales );
		gd.addNumericField( "x max :" , x_max , ncNb_decimales );
		gd.addNumericField( "y min :" , y_min , ncNb_decimales );
		gd.addNumericField( "y max :" , y_max , ncNb_decimales );
		gd.addCheckbox("Invert normals", false );
		gd.showDialog();
		if ( gd. wasCanceled ()) {
			IJ.log( "Canceled" );
			return ;
		}
		x_min = ( double )gd.getNextNumber();
		x_max = ( double )gd.getNextNumber();
		y_min = ( double )gd.getNextNumber();
		y_max = ( double )gd.getNextNumber();
		Vector vChecks = gd.getCheckboxes();
		Checkbox check_invert = (Checkbox)vChecks.get( 0 );
		boolean bInverser_normales = check_invert.getState();
		IJ.log("x_min = " + x_min + ", x_max = " + x_max );
		IJ.log("y_min = " + y_min + ", y_max = " + y_max );
		IJ.log("Invert normals = " + bInverser_normales );
		int nInverser_normales = 1;
		if ( bInverser_normales ) {
			nInverser_normales = -1;
		}

		// Création de l'image des normales de taille de l'image stack
		int nLargeur_pix = imp.getWidth();
		int nHauteur_pix = imp.getHeight();
		final int ncNbSclices = 1;
		// "RGB Black" "8-bit White"
		ImagePlus impNormales = IJ.createImage( "Normal map", 
				                                "RGB Black" , 
												nLargeur_pix, 
												nHauteur_pix, ncNbSclices );
		ImageProcessor ipNormales = impNormales.getProcessor();
		impNormales.show();
		
		// Parcours de l'image
		ImageStack imgstack = imp.getImageStack();
		Matrix M = new Matrix( nNb_lignes, 6, 0.0 );
		Matrix b = new Matrix( nNb_lignes, 1, 0.0 );
		double tab_zeros[][] = {{0. } ,{0. } ,{0.} ,{0.} ,{0.} ,{0.}};
		Matrix x = new Matrix( tab_zeros );
		final int deux_puiss_8 = (int)Math.pow(2, 8);
		final int deux_puiss_16 = (int)Math.pow(2, 16);
		int nProgression_pourcent_act = -1;
		impNormales.lockSilently();
		long tcalculStart = System.currentTimeMillis();
		for ( int v = 0; v < nHauteur_pix; ++v ) {// nHauteur_pix
			int nProgression_pourcent = (int)Math.floor( 100.0 * (double)v / (double)nHauteur_pix );
			if ( ( nProgression_pourcent_act != nProgression_pourcent ) ) {
				nProgression_pourcent_act = nProgression_pourcent;
				IJ.showProgress ( v , nHauteur_pix-2 ) ;
				impNormales.repaintWindow();
			}
			double a0, a1, a2, a3, a4, a5;
			a0 = a1 = a2 = a3 = a4 = a5 = 0;
			int nRetour = 0;
			for ( int u = 0; u < nLargeur_pix; ++u ) {// nLargeur_pix
				// calcul de la position du point A dans le repère Oijk reel
				//  bas-gauche de l'image
				double xA = ((double)u / (double)nLargeur_pix) * ( x_max - x_min ) + x_min;
				double yA = ((double)v / (double)nHauteur_pix) * ( y_max - y_min ) + y_min;
				//IJ.log( "A ( " + xA + ", " + yA + ")");
				for ( int k = 0; k < nNb_slices; k++ ) {// boucle sur la position des lampes
					// Calculer V = L - A
					double xL = vPos_lampes.get( k, 0 );
					double yL = vPos_lampes.get( k, 1 );
					//IJ.log( "L ( " + xL + ", " + yL + ")");
					double vx = xL - xA;
					double vy = yL - yA;
					double nLg_v = Math.sqrt( vx * vx + vy * vy );
					double lu = vx / nLg_v;
					double lv = vy / nLg_v;
					
					//if ( nProgression_pourcent >= 50 ) {
					//	IJ.log("Boucle k = " + k );
					//}
					
					// Remplissage du système linéaire
					M.set( k, 0, lu * lu );
					M.set( k, 1, lv * lv );
					M.set( k, 2, lu * lv );
					M.set( k, 3, lu );
					M.set( k, 4, lv );
					M.set( k, 5, 1 );
					
					ImageProcessor ip = imgstack.getProcessor( k + 1 );
					int val_pix = ip.getPixel( u, v );
					b.set( k, 0, val_pix );
				}
				// Résolution du système "au mieux"
				//IJ.log( "M = ");
				//CLinAlg2.afficher_matrice_dans_log( M );
				//IJ.log( "b = ");
				//CLinAlg2.afficher_matrice_dans_log( b );
				//IJ.log( "nRetour = " + nRetour);
				nRetour = CLinAlg.lsq(M, b, x);

				// Calcul de la normale 
				a0 = x.get(0, 0);
				a1 = x.get(1, 0);
				a2 = x.get(2, 0);
				a3 = x.get(3, 0);
				a4 = x.get(4, 0);
				a5 = x.get(5, 0);

				double den = 4.0 * a0 * a1 - a2 * a2;
				double Ix = ( a2 * a4 - 2.0 * a1 * a3 ) / den;
				double Iy = ( a2 * a3 - 2.0 * a0 * a4 ) / den;
				double Iz = Math.sqrt( 1.0 - Ix*Ix - Iy*Iy );
				//IJ.log( "N ( " + Ix + ", " + Iy + ", " + Iz + ")");
				
				// Calcul de la normale "pixel" et mise à jour de la carte des normales
				int red = (int)(nInverser_normales * Ix * 128 + 128);// - pour inverser
				int green = (int)(nInverser_normales * Iy * 128 + 128);// - pour inverser
				int blue = (int)(Iz * 255.0);
				//IJ.log( "[u = " + u + ", v = " + v + "] : (" + red + " " + green + " "+ blue + ")" );
				int nVal_pixel = red * deux_puiss_16 + 
				                 green * deux_puiss_8 +
				                 blue;
				ipNormales.putPixel(u, v, nVal_pixel );
				
			}
		}
		impNormales.unlock();
		impNormales.show();
		impNormales.repaintWindow();
		long tcalculStop = System.currentTimeMillis();
		long nDuree_secondes = ( tcalculStop - tcalculStart ) / 1000;
		IJ.log("Elapsed time : " + nDuree_secondes + " seconds" );
		IJ.log("------- Normal_map_computing_using_PTM -- end ------");
	}
}// Normal_map_computing_using_PTM
